/* WordBox:  Word search game for USR Pilot!
 *
 * (c) 1997 Kent L. Smotherman
 */

#pragma pack(2)

#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>
#include <Graffiti.h>

#include "wordbox.h"

int AppHandleEvent(EventType *);
int FormHandleEvent(EventType *);
void LoadFormVars(Word id, FormPtr form);
void InitForm(Word id, FormPtr form);
void Shutdown();

#define AppType		'Dict'			// type for application.  must be 4 chars, mixed case.
#define DBType		'Data'			// type for application database.  must be 4 chars, mixed case.
#define DBName		"Dictionary"		// name for application database.  up to 31 characters.
#define HighScoresDB  "WordBox High Scores"
#define SaveDB  "WordBox Save Game"

#define CELLWIDTH 12

typedef char* (PNames[4][100]);

static DmOpenRef	dictionary=0;
static DmOpenRef    highscoreref=0;
static DmOpenRef	saveref=0;
static LocalID      highscoreid=0;
static LocalID		saveid=0;
static char lookup[26];
static char buffer[28];
static ULong starttime=0;
static ULong gametime=240;
static Long gametimeleft=0;
static char board[26];
static int flags[25];
static SWord startx=-1,starty=-1;
static SWord index=0;
static int lastindex=-1;
static int numplayers;
static VoidPtr  Record0ptr=0;
static VoidHand Record0hand=0;
static ULong Record0size=0;
static char lastword[40];
static char curscore[20];
static char curscore2[20];
static char curtime[6];
static char curplayername[4][20];
static char *dumbptrs[4];
static Boolean replaylast = false;
static VoidHand wordlisthand[4];
static unsigned char *wordlistptr[4];
static unsigned char *wordlistwalk[4];
static long scores[4];
static UInt historyplayers=0;
static UInt uniquescoring=0;
static UInt expscoring=1;
static int curplayer=0;
static int gameinprogress=0;
static int pausemode=0;
static char *words[150];
static char *words2[150];
static VoidHand HSHand=0;
static VoidHand SaveHand=0;
static VoidPtr HSPtr=0;
static VoidPtr SavePtr=0;
static VoidHand IconHandle=0;
static VoidPtr IconPtr=0;
static VoidHand IconHandle2=0;
static VoidPtr IconPtr2=0;
static VoidHand PLists=0;
static PNames *PPtrs;
static RectangleType cliprect;
//static char *history[100];
static int sound;
static int badwordcount;
static int goodwordcount;
static int startup=0;
static char Cubes[][25] =	/* 25 6-sided cubes */
//should really be string resources but....
{
	"QBZJXK",	/*  0 */
	"HHLRDO",	/*  1 */
	"TELPCI",	/*  2 */
	"TTOTEM",   /*  3 */
	"AEAEEE",	/*  4 */
	"TOUOTO",	/*  5 */
	"NHDTHO",	/*  6 */
	"SSNSEU",	/*  7 */
	"SCTIEP",	/*  8 */
	"YIFPSR",	/*  9 */
	"OVWRGR",	/* 10 */
	"LHNROD",	/* 11 */
	"RIYPRH",	/* 12 */
	"EANDNN",	/* 13 */
	"EEEEMA",	/* 14 */
	"AAAFSR",	/* 15 */
	"AFAISR",	/* 16 */
	"DORDLN",	/* 17 */
	"MNNEAG",	/* 18 */
	"ITITIE",	/* 19 */
	"AUMEEG",	/* 20 */
	"YIFASR",	/* 21 */
	"CCWNST",	/* 22 */
	"UOTOWN",	/* 23 */
	"ETILIC"	/* 24 */
};
static long expscorevalues[] = {
    1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,16384,32768,65536,
    65536,65536,65536,65536,65536
};

static long linearscorevalues[] = {
    1,2,3,5,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11
};

static struct sNameEntry {
    void *pOnePlayer;
    void *pTwoPlayers;
    void *pThreePlayers;
    void *pFourPlayers;
    void *pPlayer1;
    void *pPlayer2;
    void *pPlayer3;
    void *pPlayer4;
    void *pNewPlayer1;
    void *pNewPlayer2;
    void *pNewPlayer3;
    void *pNewPlayer4;
    void *pPilotList41;
    void *pPilotList42;
    void *pPilotList43;
    void *pPilotList44;
    void *pNewOK;
    void *pNewCancel;
    void *pGameTime;
    void *pLinearScoring;
    void *pExpScoring;
    void *pAllScoring;
    void *pUniqueScoring;
} pNameEntry;

static struct sWordBox {
    void *pTimeLeft;
    void *pLastWord;
    void *pPause;
    void *pCancel;
    void *pScore;
    void *pName;
    void *pAccept;
    void *pWordsButton;
    void *pNewButton;
} pWordBox;

static struct sWords {
    void *pClose;
    void *pPlayerNames;
    void *pPlayerScore;
    void *pPlayerWords;
    void *pPlayerNameList;
    void *pPlayerNames2;
    void *pPlayerScore2;
    void *pPlayerWords2;
    void *pPlayerName2List;
    void *pHSButton;
	void *pOptions;
} pWords;

static struct {
    void *pHSGameTime;
    void *pHSGameTimeList;
    void *pHSPilotScoring;
    void *pHSExpScoring;
    void *pHSLinearScoring;
    void *pHSAllScoring;
    void *pHSUniqueScoring;
    void *pHSList;
    void *pHSClose;
    void *pHighScores;
    void *pHSClear;
    void *pOnePlayer;
    void *pTwoPlayers;
    void *pThreePlayers;
    void *pFourPlayers;
} pHighScores;

typedef struct {
    UInt numplayers;
    UInt linearscoring;
    UInt uniquescoring;
    ULong gametime;
    UInt historyplayers;
    unsigned char names[100][20];
    int archiveorder[4][100];
} tWordBoxPrefs;

typedef struct {
    UInt numscores;
} tHighScoreHdr;

typedef struct {
    ULong score;
    char name[20];
    char date[dateStringLength];
} tHighScoreItem;

typedef struct {
	int gameinprogress;
	char names[4][20];
	UInt curplayer;
	char board[26];
	int flags[25];
	Long gametimeleft;
	long scores[4];
	int goodwordcount;
	int badwordcount;
	long wordsize[4];
} tSaveHdr;

static VoidHand prefhandle=0;
static tWordBoxPrefs *prefptr=0;
static char *dummystrings[10];

static SndCommandType sndcmd,sndcmd2,sndcmd3;
void GameOver();

int FindWord(char *word) {
    unsigned char *temp = wordlistptr[curplayer];
    unsigned char *walk,*copy;
    unsigned char local[8];
    int i,j,rc;
    while (temp < wordlistwalk[curplayer]) {
        rc = StrCompare(word,temp+7);
        if (!rc)
            return 1;
        if (rc < 0) { //found the insert point
            break;
        }
        temp += StrLen(temp)+1;
    }
    walk = wordlistwalk[curplayer]-1;
    copy = walk+StrLen(word)+1+7;
    while (walk >= temp) {
        *copy = *walk;
        walk--;
        copy--;
    }
    wordlistwalk[curplayer] += StrLen(word)+1+7;
    if (prefptr->linearscoring)
        StrIToA(local,linearscorevalues[StrLen(word)-4]);
    else
        StrIToA(local,expscorevalues[StrLen(word)-4]);
    for (i=0;i<(5-StrLen(local));i++)
        temp[i] = ' ';
    for (j=0;i<5;i++) {
        temp[i] = local[j];
        j++;
    }
    temp[5] = ' ';
    temp[6] = ' ';
    StrCopy(temp+7,word);
    return 0;
}

int VerifyWord(char *word) {
    VoidHand hand;
    VoidPtr ptr;
    ULong size;
    UInt rec;
    ULong hash;
    int i,l,rc,count=0;
    unsigned char *scan;

    rc = 0;
    if (StrLen(word) < 4 || !gameinprogress)
        return 0;
    hash = (word[0]-'a')*676 + (word[1]-'a')*26 + (word[2]-'a');
    rec = 1;
    count = 0;
    scan = Record0ptr;
    while (count < Record0size) {
        if (word[0]==scan[0])
            if (word[1]==scan[1])
                if (word[2]==scan[2])
                    break;
                else {
                    if (word[2]<scan[2]) return 0;
                }
             else {
                if (word[1]<scan[1]) return 0;
             }
         else {
            if (word[0]<scan[0]) return 0;
         }
         rec++;
         count += 3;
         scan += 3;
    }
    if (count >= Record0size)
        return 0;
    if (hand = DmQueryRecord(dictionary,rec)) {
        ptr = MemHandleLock(hand);
        size = MemHandleSize(hand);
        scan = ptr;
        count = 0;
        l = 0;
        while (count < size) {
            i = (*scan)/32;
            if (i != 0 && count != 0) {
                buffer[l] = 0;
                rc = StrCompare(word,buffer);
                if (rc < 0) {
                    rc = 0;
                    break;
                }
                if (rc == 0) {
                    rc = 1;
                    break;
                }
                rc = 0;
                if (i == 7)
                    l = 0;
                else
                    l = i+2;
            }
            buffer[l] = ((*scan)&0x1f)+'a';
            l++;
            scan++;
            count++;
        }
        buffer[l] = 0;
        rc = StrCompare(word,buffer);
        if (rc < 0) {
            rc = 0;
        }
        else if (rc == 0) {
            rc = 1;
        }
        else rc = 0;
        MemHandleUnlock(hand);
    }
    if (rc == 1) {
        rc = FindWord(word);
        if (rc == 0)
            rc = 1;
        else
            rc = 2;
    }
    return rc;
}

ULong ScorePlayerWordList(int player, UInt *idx,char **words, int p2) {
    UInt i,j,rc,rrc;
    long score;
    unsigned char *temp;
    unsigned char *walk;

    if (!wordlisthand[player]) return 0;
    temp = wordlistptr[player];
    i = 0;
    score = 0;
    while (temp < wordlistwalk[player]) {
		rrc = 0;
        words[i] = (char *)temp;
		for (j=0;j<numplayers;j++) {
            if (j==player) continue;
            walk = wordlistptr[j];
		    rc = 1;
            while (walk < wordlistwalk[j]) {
                rc = StrCompare(temp+7,walk+7);
                if (rc <= 0)
                    break;
                walk += StrLen(walk)+1;
            }
            if (rc == 0) {
				if ((j != p2)) {
					rrc = 1;
				}
				else {
					rrc = 2;
					break;
				}
			}
        }
        if (!rrc) {
            temp[0] = '*';
            if (prefptr->linearscoring)
                score += linearscorevalues[StrLen(temp+7)-4];
            else
                score += expscorevalues[StrLen(temp+7)-4];
        }
        else {
			if (rrc == 1)
				temp[0] = '-';
			else
				temp[0] = ' ';
            if (!prefptr->uniquescoring) {
                if (prefptr->linearscoring)
                    score += linearscorevalues[StrLen(temp+7)-4];
                else
                    score += expscorevalues[StrLen(temp+7)-4];
                }
        }
        temp += StrLen(temp)+1;
        i++;
    }
    *idx = i;
    return score;
}

void BuildWordList() {
    UInt sel1,sel2,i;
    long score;
	char temp[10];
	int scoring;

	LstEraseList(pWords.pPlayerWords2);
	LstEraseList(pWords.pPlayerWords);
    scoring = prefptr->uniquescoring;
	prefptr->uniquescoring = 1;
	sel1 = LstGetSelection(pWords.pPlayerNameList);
    sel2 = LstGetSelection(pWords.pPlayerName2List);
    if (sel1 < 0 || (numplayers > 1 && sel2 < 0)) return;
    score = ScorePlayerWordList(sel1,&i,words,sel2);
    if (i<9)
		LstSetHeight(pWords.pPlayerWords,i);
	else
		LstSetHeight(pWords.pPlayerWords,9);
	LstSetListChoices(pWords.pPlayerWords,words,i);
    LstMakeItemVisible(pWords.pPlayerWords,0);
    LstDrawList(pWords.pPlayerWords);
    StrIToA(curscore,score);
    StrIToA(temp,scores[sel1]);
	StrCat(curscore,"/");
	StrCat(curscore,temp);
    FldSetTextPtr(pWords.pPlayerScore,curscore);
    FldEraseField(pWords.pPlayerScore);
    FldDrawField(pWords.pPlayerScore);

    score = ScorePlayerWordList(sel2,&i,words2,sel1);
    if (i<9)
		LstSetHeight(pWords.pPlayerWords2,i);
	else
		LstSetHeight(pWords.pPlayerWords2,9);
    LstSetListChoices(pWords.pPlayerWords2,words2,i);
    LstMakeItemVisible(pWords.pPlayerWords2,0);
    LstDrawList(pWords.pPlayerWords2);
    StrIToA(curscore2,score);
    StrIToA(temp,scores[sel2]);
	StrCat(curscore2,"/");
	StrCat(curscore2,temp);
    FldSetTextPtr(pWords.pPlayerScore2,curscore2);
    FldEraseField(pWords.pPlayerScore2);
    FldDrawField(pWords.pPlayerScore2);
    prefptr->uniquescoring = scoring;
}

void SaveAppPrefs() {
    int i,j;
	VoidHand rec0hand;
	VoidHand hand;
	char	*ptr;
	tSaveHdr *rec0ptr;
	UInt rec=0;
	ULong count;

    if (!prefhandle) return;
    prefptr->numplayers = numplayers;
    prefptr->gametime = gametime;
    prefptr->historyplayers = historyplayers;
	PrefSetAppPreferences ('WBOX', 2, prefptr, sizeof(tWordBoxPrefs));
	if (saveref) {
		rec0hand = DmGetRecord(saveref,rec);
		if (!rec0hand) return;
		rec0ptr = MemHandleLock(rec0hand);
        DmWrite(rec0ptr,(ULong)&(rec0ptr->gameinprogress)-(ULong)rec0ptr,&gameinprogress,sizeof(UInt));
		if (wordlistptr[0]) {
	        DmWrite(rec0ptr,(ULong)&(rec0ptr->gametimeleft)-(ULong)rec0ptr,&gametimeleft,sizeof(long));
			pausemode = 1;
	        DmWrite(rec0ptr,(ULong)&(rec0ptr->curplayer)-(ULong)rec0ptr,&curplayer,sizeof(UInt));
	        DmWrite(rec0ptr,(ULong)&(rec0ptr->gameinprogress)-(ULong)rec0ptr,&gameinprogress,sizeof(UInt));
	        DmStrCopy(rec0ptr,(ULong)&(rec0ptr->board)-(ULong)rec0ptr,board);
	        DmWrite(rec0ptr,(ULong)(rec0ptr->flags)-(ULong)rec0ptr,flags,sizeof(UInt)*25);
	        DmWrite(rec0ptr,(ULong)&(rec0ptr->goodwordcount)-(ULong)rec0ptr,&goodwordcount,sizeof(UInt));
	        DmWrite(rec0ptr,(ULong)&(rec0ptr->badwordcount)-(ULong)rec0ptr,&badwordcount,sizeof(UInt));
	        DmWrite(rec0ptr,(ULong)(rec0ptr->scores)-(ULong)rec0ptr,scores,sizeof(long)*4);
			for (i=0;i<numplayers;i++)
				DmStrCopy(rec0ptr,(ULong)(rec0ptr->names[i])-(ULong)rec0ptr,curplayername[i]);
			for (i=0;i<numplayers;i++) {
		        count = (ULong)wordlistwalk[i]-(ULong)wordlistptr[i];
				hand = DmGetRecord(saveref,i+1);
				ptr = MemHandleLock(hand);
				DmWrite(rec0ptr,(ULong)&(rec0ptr->wordsize[i])-(ULong)rec0ptr,&count,sizeof(ULong));
				DmWrite(ptr,0,wordlistptr[i],4000);
				MemHandleUnlock(hand);
		        DmReleaseRecord(saveref,i+1,true);
			}
		}
		MemHandleUnlock(rec0hand);
		rec = 0;
        DmReleaseRecord(saveref,rec,true);
	}
}

void LoadSavedGame(int ignoreinprogress) {
	VoidHand rec0hand;
	tSaveHdr *rec0ptr;
	UInt rec=0;
	VoidHand hand;
	char *ptr;
	int i;

	if (saveref) {
		rec0hand = DmGetRecord(saveref,rec);
		if (!rec0hand) return;
		rec0ptr = MemHandleLock(rec0hand);
		gametimeleft = rec0ptr->gametimeleft;
		curplayer = rec0ptr->curplayer;
		if (ignoreinprogress) {
			StrCopy(board,rec0ptr->board);
		}
		else if (rec0ptr->gameinprogress) {
			pausemode = 1;
			startup = 2;
			StrCopy(board,rec0ptr->board);
			for (i=0;i<25;i++)
				flags[i] = rec0ptr->flags[i];
		}
		else {
			pausemode = 0;
			startup = 0;
		}
		goodwordcount = rec0ptr->goodwordcount;
		badwordcount = rec0ptr->badwordcount;
		for (i=0;i<numplayers;i++) {
			StrCopy(curplayername[i],rec0ptr->names[i]);
			scores[i] = rec0ptr->scores[i];
		}
		for (i=0;i<numplayers;i++) {
			hand = DmGetRecord(saveref,i+1);
			ptr = MemHandleLock(hand);
			if (wordlisthand[i] == 0) {
				wordlisthand[i] = MemHandleNew(4000);
				ErrFatalDisplayIf(!wordlisthand[i],"Could not allocate words memory");
    	    	wordlistptr[i] = MemHandleLock(wordlisthand[i]);
			}
			MemMove(wordlistptr[i],ptr,4000);
			wordlistwalk[i] = wordlistptr[i] + rec0ptr->wordsize[i];
			MemHandleUnlock(hand);
		    DmReleaseRecord(saveref,i+1,false);
		}
		MemHandleUnlock(rec0hand);
		rec = 0;
        DmReleaseRecord(saveref,rec,false);
	}
}

void GetAppPrefs() {
    int i,j;

    prefhandle = MemHandleNew(sizeof(tWordBoxPrefs));
    if (!prefhandle) return;
    prefptr = MemHandleLock(prefhandle);
    MemSet(prefptr,sizeof (tWordBoxPrefs),0);
	if (PrefGetAppPreferences ('WBOX', 2, prefptr, sizeof(tWordBoxPrefs))) {
        numplayers = prefptr->numplayers;
        gametime = prefptr->gametime;
        historyplayers = prefptr->historyplayers;
//        for (i=0;i<historyplayers;i++)
//            history[i] = prefptr->names[i];
    }
    else {
        for (i=0;i<4;i++)
            for (j=0;j<100;j++)
                prefptr->archiveorder[i][j] = j;
    }
	LoadSavedGame(0);
}

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    short err;
    EventType e;
    void *display;
    void *form,*text,*list;
    int i,count,l;
    char *c,*t;
    UInt rec;
    VoidHand hand;

    if (!cmd) {
		if (!PLists) {
			PLists = MemHandleNew(sizeof(PNames));
			ErrFatalDisplayIf(!PLists,"Could not allocate players names memory");
    		PPtrs = MemHandleLock(PLists);
		}
        highscoreid = DmFindDatabase(0,HighScoresDB);
        if (!highscoreid) {
            DmCreateDatabase(0,HighScoresDB,'WBOX','HIGH',false);
            highscoreid = DmFindDatabase(0,HighScoresDB);
            highscoreref = DmOpenDatabase(0,highscoreid,dmModeReadWrite);
            for (i=0;i<40;i++) {
                rec = i;
                hand = DmNewRecord(highscoreref,&rec,(ULong)(sizeof(tHighScoreHdr)+
                    sizeof(tHighScoreItem)*10));
                c = MemHandleLock(hand);
                DmSet(c,0,(ULong)(sizeof(tHighScoreHdr)+
                    sizeof(tHighScoreItem)*10),0);
                MemHandleUnlock(hand);
                DmReleaseRecord(highscoreref,rec,true);
            }
        }
        else
            highscoreref = DmOpenDatabase(0,highscoreid,dmModeReadWrite);
        saveid = DmFindDatabase(0,SaveDB);
		if (!saveid) {
			DmCreateDatabase(0,SaveDB,'WBOX','SAVE',false);
			saveid = DmFindDatabase(0,SaveDB);
	        saveref = DmOpenDatabase(0,saveid,dmModeReadWrite);
            rec = 0;
			hand = DmNewRecord(saveref,&rec,(ULong)(sizeof(tSaveHdr)));
            c = MemHandleLock(hand);
            DmSet(c,0,(ULong)sizeof(tSaveHdr),0);
            MemHandleUnlock(hand);
            DmReleaseRecord(saveref,rec,true);
			for (i=1;i<=4;i++) {
                rec = i;
                hand = DmNewRecord(saveref,&rec,(ULong)(4000));
                c = MemHandleLock(hand);
                DmSet(c,0,(ULong)4000,0);
                MemHandleUnlock(hand);
                DmReleaseRecord(saveref,rec,true);
			}
		}
		else {
	        saveref = DmOpenDatabase(0,saveid,dmModeReadWrite);
		}
        for (i=0;i<10;i++)
            dummystrings[i] = "1";
        *(curplayername[0]) = 0;
        *(curplayername[1]) = 0;
        *(curplayername[2]) = 0;
        *(curplayername[3]) = 0;
  	    cliprect.extent.x = 0;
  	    sndcmd.cmd = sndCmdFreqDurationAmp;
  	    sndcmd.param1 = 600;
  	    sndcmd.param2 = 100;
  	    sndcmd.param3 = sndMaxAmp/4;
  	    sndcmd2.cmd = sndCmdFreqDurationAmp;
  	    sndcmd2.param1 = 800;
  	    sndcmd2.param2 = 100;
  	    sndcmd2.param3 = sndMaxAmp/4;
  	    sndcmd3.cmd = sndCmdFreqDurationAmp;
  	    sndcmd3.param1 = 1000;
  	    sndcmd3.param2 = 100;
  	    sndcmd3.param3 = sndMaxAmp/4;
        for (i=0;i<4;i++)
            wordlisthand[i] = 0;
  	    starttime = TimGetTicks();
        *lastword = 0;
        *curtime = 0;
        *curscore = 0;
        dictionary = DmOpenDatabaseByTypeCreator(DBType, AppType, dmModeReadOnly);
        Record0hand = DmQueryRecord(dictionary,0);
        Record0ptr = MemHandleLock(Record0hand);
        Record0size = MemHandleSize(Record0hand);
        IconHandle = DmGetResource(bitmapRsc, 1000);
        ErrFatalDisplayIf(!IconHandle, "Unable to find bitmap handle");
        IconPtr = MemHandleLock(IconHandle);
        ErrFatalDisplayIf(!IconPtr, "Unable to allocate bitmap memory");
        IconHandle2 = DmGetResource(bitmapRsc, 1001);
        ErrFatalDisplayIf(!IconHandle2, "Unable to find bitmap handle");
        IconPtr2 = MemHandleLock(IconHandle2);
        ErrFatalDisplayIf(!IconPtr2, "Unable to allocate bitmap memory");

        StrCopy(board,"WORDQAFRDBYIACORSDNXORUSM");
        for (i=0;i<25;i++)
            flags[0] = 0;
        flags[0] = 1;
        flags[1] = 1;
        flags[2] = 1;
        flags[3] = 1;
        flags[9] = 1;
        flags[14] = 1;
        flags[19] = 1;
		startup = 0;
		gameinprogress = 0;

        GetAppPrefs();	//must be after init of global data above so
						// restored game can overrite that data

        FrmGotoForm(WordBox);
        FrmUpdateForm(WordBox,frmRedrawUpdateCode);

        while(1) {
            EvtGetEvent(&e, 100);   //cause nilEvent every second
            if (SysHandleEvent(&e)) continue;
            if (MenuHandleEvent((void *)0, &e, &err)) continue;
            if (AppHandleEvent(&e)) continue;
            if (FormHandleEvent(&e)) continue;
            if (e.eType == appStopEvent) {
                Shutdown();
                break;
            }
            else
                FrmHandleEvent(FrmGetActiveForm(), &e);
        }
    }
}

void Shutdown() {
    int i;
    SaveAppPrefs();
    if (prefhandle) {
        MemHandleUnlock(prefhandle);
        MemHandleFree(prefhandle);
    }
    FrmCloseAllForms();
    if (Record0hand)
        MemHandleUnlock(Record0hand);
    if (dictionary)
        DmCloseDatabase(dictionary);
    if (highscoreref)
        DmCloseDatabase(highscoreref);
    if (saveref)
        DmCloseDatabase(saveref);
    MemPtrUnlock(IconPtr);
    DmReleaseResource(IconHandle);
    MemPtrUnlock(IconPtr2);
    DmReleaseResource(IconHandle2);

    dictionary = 0;
    prefhandle = 0;
    prefptr = 0;
    Record0hand = 0;
    IconPtr = 0;
    IconPtr2 = 0;
    IconHandle = 0;
    IconHandle2 = 0;
    return;
}

int AppHandleEvent(EventType *e) {
    FormPtr form;

    switch (e->eType) {
        case frmLoadEvent:
            form = FrmInitForm(e->data.frmLoad.formID);
            FrmSetActiveForm(form);
            LoadFormVars(e->data.frmLoad.formID,form);
            return true;
        case frmOpenEvent:
			InitForm(e->data.frmLoad.formID,form);
            FrmDrawForm(FrmGetActiveForm());
            return true;
    }
    return false;
}

void GameOver() {
    int i;

    index = 0;
    gameinprogress = 0;
/*
    for (i=0;i<numplayers;i++) {
        if (wordlisthand[i]) {
            MemHandleUnlock(wordlisthand[i]);
            MemHandleFree(wordlisthand[i]);
            wordlisthand[i] = 0;
            wordlistptr[i] = 0;
            wordlistwalk[i] = 0;
        }
    }
*/
}

void UpdateHighScores(Boolean reset) {
    int i,j,k;
    ULong score=0;
    VoidHand hand;
    char *ptr;
    ULong size;
    UInt rec,dummy;
    Boolean dirty;
    UInt num;
    Err rc;
    tHighScoreItem *item,*walk;
    tHighScoreHdr *hdr;
    DateTimeType date;

    dirty = false;
	if (prefptr->uniquescoring) {
	    rec = (gametime/60-2)*8+8;
		rec += (numplayers-1)*2;
		if (prefptr->linearscoring)
			rec += 1;
	}
	else {
	    rec = (gametime/60-2)*2;
	    if (prefptr->linearscoring)
		    rec += 1;
	}
    hand = DmGetRecord(highscoreref,rec);
    ErrFatalDisplayIf(!hand,"No handle to High Scores DB");
    ptr = MemHandleLock(hand);
    ErrFatalDisplayIf(!ptr,"No ptr to High Scores DB");
    size = MemHandleSize(hand);
    hdr = (tHighScoreHdr *)ptr;
    if (reset) {
        num = 0;
        DmWrite(ptr,(ULong)&(hdr->numscores)-(ULong)ptr,&num,sizeof(UInt));
        MemHandleUnlock(hand);
        rc = DmReleaseRecord(highscoreref,rec,dirty);
        return;
    }
    num = hdr->numscores;
    item = (tHighScoreItem *)(ptr+sizeof(tHighScoreHdr));
    TimSecondsToDateTime(TimGetSeconds(),&date);
    for (i=0;i<numplayers;i++) {
        score = ScorePlayerWordList(i,&dummy,words,-1);
        walk = item;
        for (j=0;j<num;j++) {
            if (walk->score < score)
                break;
            walk++;
        }
        if (j < 10) {
            for (k=8;k>=j;k--) {
                DmWrite(ptr,(ULong)&item[k+1]-(ULong)ptr,&item[k],sizeof(tHighScoreItem));
            }
            rc = DmWrite(ptr,(ULong)walk->name-(ULong)ptr,curplayername[i],StrLen(curplayername[i])+1);
            ErrFatalDisplayIf(rc,"HighScore DB write error");
            DmWrite(ptr,(ULong)&(walk->score)-(ULong)ptr,&score,sizeof(ULong));
            if (num < 10) {
                num++;
                rc = DmWrite(ptr,(ULong)&(hdr->numscores)-(ULong)ptr,&num,sizeof(UInt));
                ErrFatalDisplayIf(rc,"HighScore DB write error");
            }
            DateToAscii(date.month,date.day,date.year,dfMDYWithSlashes,buffer);
            DmWrite(ptr,(ULong)walk->date-(ULong)ptr,buffer,StrLen(buffer)+1);
            dirty = true;
        }
    }
    MemHandleUnlock(hand);
    rc = DmReleaseRecord(highscoreref,rec,dirty);
    ErrFatalDisplayIf(rc,"HighScore DB release error");
}

void UpdateClock() {
    Long mytime;
    int mins,secs,i;
    char temp[5];

    if (gameinprogress == 0 || pausemode)
        return;
    gametimeleft -= (TimGetTicks()-starttime);    //num seconds remaining in game
    starttime = TimGetTicks();
    mytime = gametimeleft/100;
    if (mytime < 0 || gametimeleft < 0) {
        for (i=0;i<25;i++)
            flags[i] = 0;
        cliprect.extent.x = 0;
        *lastword = 0;
		index = 0;
        StrCopy(curtime,"0:00");
        FldSetTextPtr(pWordBox.pTimeLeft,curtime);
        FldEraseField(pWordBox.pTimeLeft);
        FldDrawField(pWordBox.pTimeLeft);
        FldSetTextPtr(pWordBox.pLastWord,lastword);
        FldEraseField(pWordBox.pLastWord);
        FldDrawField(pWordBox.pLastWord);
        gameinprogress = 0;
        curplayer++;
        SndDoCmd(0,&sndcmd2,false);
        SndDoCmd(0,&sndcmd,false);
        if (curplayer >= numplayers) {
            curplayer = 0;
            UpdateHighScores(false);
            FrmGotoForm(Words);
            return;
        }
        goodwordcount = 0;
        badwordcount = 0;
        pausemode = 1;
        index = 0;
	    starttime = TimGetTicks();
	    gametimeleft = gametime*100;
		mytime = gametimeleft/100;
	    gameinprogress = 1;
        FldSetTextPtr(pWordBox.pName,curplayername[curplayer]);
        FldEraseField(pWordBox.pName);
        FldDrawField(pWordBox.pName);
        scores[curplayer] = 0;
        StrCopy(curscore,"0");
        StrCopy(lastword,"");
        FldSetTextPtr(pWordBox.pScore,curscore);
        FldSetTextPtr(pWordBox.pLastWord,lastword);
        FldEraseField(pWordBox.pScore);
        FldDrawField(pWordBox.pScore);
        FldEraseField(pWordBox.pLastWord);
        FldDrawField(pWordBox.pLastWord);
        FrmUpdateForm(WordBox,frmRedrawUpdateCode);
//        UpdateClock();
//        return;
    }
    mins = (int)mytime/60;
    secs = ((int)mytime)%60;
    curtime[0] = mins+'0';
    curtime[1] = ':';
    curtime[2] = (secs/10)+'0';
    curtime[3] = (secs%10)+'0';
    curtime[4] = 0;
    FldSetTextPtr(pWordBox.pTimeLeft,curtime);
    FldEraseField(pWordBox.pTimeLeft);
    FldDrawField(pWordBox.pTimeLeft);
}

void GenerateGame() {
    int i,j;
    char c;

    badwordcount = 0;
    goodwordcount = 0;
    if (!replaylast) {
        SysRandom(TimGetSeconds());
        for (i=0;i<25;i++)
            board[i] = i;
        for (i=0;i<25;i++) {
            j = SysRandom(0)%25;
            c = board[i];
            board[i] = board[j];
            board[j] = c;
            flags[i] = 0;
        }
        for (i=0;i<25;i++) {
            j = SysRandom(0)%6;
            board[i] = Cubes[board[i]][j];
        }
        board[25] = 0;
    }
    replaylast = false;
    index = 0;
    cliprect.extent.x = 0;
/*
    scores[curplayer] = 0;
    StrCopy(curscore,"0");
    StrCopy(lastword,"");
    FrmUpdateForm(WordBox,frmRedrawUpdateCode);
    FldSetTextPtr(pWordBox.pScore,curscore);
    FldSetTextPtr(pWordBox.pLastWord,lastword);
    FldEraseField(pWordBox.pScore);
    FldDrawField(pWordBox.pScore);
    FldEraseField(pWordBox.pLastWord);
    FldDrawField(pWordBox.pLastWord);
*/
}

void DrawBoard() {
    int i;
    short w;
    char c[3];
    SWord x,y;
    void *win;
    RectangleType rect;

    rect.topLeft.x = 1;
    rect.topLeft.y = 1;
    rect.extent.x = 124;
    rect.extent.y = 124;
    FntSetFont(2);
    c[1] = 0;
    c[2] = 0;
    WinSetDrawWindow(FrmGetWindowHandle(FrmGetActiveForm()));
    if (cliprect.extent.x > 0)
        WinSetClip(&cliprect);
    WinDrawRectangleFrame(dialogFrame,&rect);
    rect.topLeft.x = 2;
    rect.topLeft.y = 2;
    rect.extent.x = 120;
    rect.extent.y = 120;
    WinEraseRectangle(&rect,0);
	FntSetFont(0);
	if (prefptr->linearscoring)
		WinDrawChars("Linear",6,128,100);
	else
		WinDrawChars("Exp.",4,128,100);
	if (prefptr->uniquescoring)
		WinDrawChars("Unique",6,128,112);
	else
		WinDrawChars("All",3,128,112);
	FntSetFont(2);
    if (pausemode) {
        WinDrawChars("Tap Pause to resume",19,5,55);
        return;
    }
    for (i=0;i<25;i++) {
        c[0] = board[i];
        c[1] = 0;
        y = (i/5)*24+2;
        x = (i%5)*24+2;
        rect.topLeft.x = x+1;
        rect.topLeft.y = y+3;
        rect.extent.x = 24;
        rect.extent.y = 24;
        if (flags[i])
            WinDrawBitmap(IconPtr2, rect.topLeft.x, rect.topLeft.y);
//        else
//            WinDrawBitmap(IconPtr, x, y);
        w = FntCharWidth(c[0]);
        x += (24-w)/2;
        y += 6;
        if (flags[i])
            WinDrawInvertedChars(c,1,x,y);
        else {
            if (c[0] != 'Q')
                WinDrawChars(c,1,x,y);
            else {
                c[1] = 'u';
                WinDrawChars(c,2,x-2,y);
            }
        }
//        if (flags[i]) {
//            WinInvertRectangle(&rect,0);
//        }
    }
    WinResetClip();
}

void processStroke(EventType *e) {
    int x,y,i,dx,dy;
	dx = 0;
    if (!gameinprogress) return;
    if (e->screenX >= 2+(24-CELLWIDTH)/2 && e->screenX < 122-(24-CELLWIDTH)/2)
        if (e->screenY >= 2+(24-CELLWIDTH)/2 && e->screenY < 122-(24-CELLWIDTH)/2)
            for (y=2+(24-CELLWIDTH)/2;y<5*24;y+=24)
	            if (e->screenY >= y && e->screenY < y+CELLWIDTH)
                    for (x=2+(24-CELLWIDTH)/2;x<5*24;x+=24)
			            if (e->screenX >= x && e->screenX < x+CELLWIDTH) {
							dx = 1;
						}
	if (!dx) return;
	x = (e->screenX-2)/24;
	y = (e->screenY-2)/24;
    i = y*5+x;
    if (flags[i] != 0) return;
    if (startx != -1) {
        if (x >= startx)
            dx = x-startx;
        else
            dx = startx-x;
        if (y >= starty)
            dy = y-starty;
        else
            dy = starty-y;
        if (dx > 1 || dy > 1) return;
    }
    lookup[index] = board[i];
    index++;
    flags[i] = 1;
    cliprect.topLeft.x = x*24+2;
    cliprect.topLeft.y = y*24+2;
    cliprect.extent.x = 24;
    cliprect.extent.y = 24;
    FrmUpdateForm(WordBox,frmRedrawUpdateCode);
    startx = x;
    starty = y;
}

void ArchivePlayer(char *name,int idx) {
    int i,j,t;
    ErrFatalDisplayIf(!name,"Null name");
    if (historyplayers == 100)
        return;
    for (i=0;i<historyplayers;i++) {
        ErrFatalDisplayIf(!(prefptr->names[i]),"Null history");
        if (!StrCompare(prefptr->names[prefptr->archiveorder[idx][i]],name)) {
            break;
        }
    }

    if (i >= historyplayers) {
        StrCopy(prefptr->names[historyplayers],name);
//        history[historyplayers] = prefptr->names[historyplayers];	//5-19-97
        prefptr->archiveorder[idx][historyplayers] = historyplayers;	//5-19-97
		(*PPtrs)[idx][historyplayers] = prefptr->names[historyplayers];
        historyplayers++;
    }
    /*else*/ {	//5-19-97
        t = prefptr->archiveorder[idx][i];
        for (j=i-1;j>=0;j--) {
            prefptr->archiveorder[idx][j+1] = prefptr->archiveorder[idx][j];
			(*PPtrs)[idx][j+1] = (*PPtrs)[idx][j];
        }
        prefptr->archiveorder[idx][0] = t;
		(*PPtrs)[idx][0] = prefptr->names[t];
    }
	if (historyplayers < 4)
		LstSetHeight(pNameEntry.pPilotList41,historyplayers);
	LstSetListChoices(pNameEntry.pPilotList41,(*PPtrs)[0],historyplayers);
    if (historyplayers < 4)
		LstSetHeight(pNameEntry.pPilotList42,historyplayers);
    LstSetListChoices(pNameEntry.pPilotList42,(*PPtrs)[1],historyplayers);
    if (historyplayers < 4)
		LstSetHeight(pNameEntry.pPilotList43,historyplayers);
    LstSetListChoices(pNameEntry.pPilotList43,(*PPtrs)[2],historyplayers);
    if (historyplayers < 4)
		LstSetHeight(pNameEntry.pPilotList44,historyplayers);
    LstSetListChoices(pNameEntry.pPilotList44,(*PPtrs)[3],historyplayers);
}

void SetNumPlayers(int n) {
    switch (n) {
        case 1:
            CtlSetValue(pNameEntry.pOnePlayer,1);
            numplayers = 1;
            FldSetUsable(pNameEntry.pNewPlayer1,true);
            FldDrawField(pNameEntry.pNewPlayer1);
            CtlShowControl(pNameEntry.pPlayer1);
            FldEraseField(pNameEntry.pNewPlayer2);
            FldSetUsable(pNameEntry.pNewPlayer2,false);
            FldEraseField(pNameEntry.pNewPlayer2);
            FldSetUsable(pNameEntry.pNewPlayer2,false);
            CtlHideControl(pNameEntry.pPlayer2);
            FldEraseField(pNameEntry.pNewPlayer3);
            FldSetUsable(pNameEntry.pNewPlayer3,false);
            CtlHideControl(pNameEntry.pPlayer3);
            FldEraseField(pNameEntry.pNewPlayer4);
            FldSetUsable(pNameEntry.pNewPlayer4,false);
            CtlHideControl(pNameEntry.pPlayer4);
            break;
        case 2:
            CtlSetValue(pNameEntry.pTwoPlayers,1);
            numplayers = 2;
            FldSetUsable(pNameEntry.pNewPlayer1,true);
            FldDrawField(pNameEntry.pNewPlayer1);
            CtlShowControl(pNameEntry.pPlayer1);
            FldSetUsable(pNameEntry.pNewPlayer2,true);
            FldDrawField(pNameEntry.pNewPlayer2);
            CtlShowControl(pNameEntry.pPlayer2);
            FldEraseField(pNameEntry.pNewPlayer3);
            FldSetUsable(pNameEntry.pNewPlayer3,false);
            CtlHideControl(pNameEntry.pPlayer3);
            FldEraseField(pNameEntry.pNewPlayer4);
            FldSetUsable(pNameEntry.pNewPlayer4,false);
            CtlHideControl(pNameEntry.pPlayer4);
            break;
        case 3:
            CtlSetValue(pNameEntry.pThreePlayers,1);
            numplayers = 3;
            FldSetUsable(pNameEntry.pNewPlayer1,true);
            FldDrawField(pNameEntry.pNewPlayer1);
            CtlShowControl(pNameEntry.pPlayer1);
            FldSetUsable(pNameEntry.pNewPlayer2,true);
            FldDrawField(pNameEntry.pNewPlayer2);
            CtlShowControl(pNameEntry.pPlayer2);
            FldSetUsable(pNameEntry.pNewPlayer3,true);
            FldDrawField(pNameEntry.pNewPlayer3);
            CtlShowControl(pNameEntry.pPlayer3);
            FldEraseField(pNameEntry.pNewPlayer4);
            FldSetUsable(pNameEntry.pNewPlayer4,false);
            CtlHideControl(pNameEntry.pPlayer4);
            break;
        case 4:
            CtlSetValue(pNameEntry.pFourPlayers,1);
            numplayers = 4;
            FldSetUsable(pNameEntry.pNewPlayer1,true);
            FldDrawField(pNameEntry.pNewPlayer1);
            CtlShowControl(pNameEntry.pPlayer1);
            FldSetUsable(pNameEntry.pNewPlayer2,true);
            FldDrawField(pNameEntry.pNewPlayer2);
            CtlShowControl(pNameEntry.pPlayer2);
            FldSetUsable(pNameEntry.pNewPlayer3,true);
            FldDrawField(pNameEntry.pNewPlayer3);
            CtlShowControl(pNameEntry.pPlayer3);
            FldSetUsable(pNameEntry.pNewPlayer4,true);
            FldDrawField(pNameEntry.pNewPlayer4);
            CtlShowControl(pNameEntry.pPlayer4);
            break;
        default:
            numplayers = 1;
            FldSetUsable(pNameEntry.pNewPlayer1,false);
            FldEraseField(pNameEntry.pNewPlayer1);
            CtlHideControl(pNameEntry.pPlayer1);
            FldEraseField(pNameEntry.pNewPlayer2);
            FldSetUsable(pNameEntry.pNewPlayer2,false);
            FldEraseField(pNameEntry.pNewPlayer2);
            FldSetUsable(pNameEntry.pNewPlayer2,false);
            CtlHideControl(pNameEntry.pPlayer2);
            FldEraseField(pNameEntry.pNewPlayer3);
            FldSetUsable(pNameEntry.pNewPlayer3,false);
            CtlHideControl(pNameEntry.pPlayer3);
            FldEraseField(pNameEntry.pNewPlayer4);
            FldSetUsable(pNameEntry.pNewPlayer4,false);
            CtlHideControl(pNameEntry.pPlayer4);
            break;
    }
}

void DisplayHighScores() {
    UInt rec;
	VoidPtr form;
	int num;
	
    if (CtlGetValue(pHighScores.pOnePlayer))
		num = 0;
	else if (CtlGetValue(pHighScores.pTwoPlayers))
		num = 1;
	else if (CtlGetValue(pHighScores.pThreePlayers))
		num = 2;
	else if (CtlGetValue(pHighScores.pFourPlayers))
		num = 3;
	LstEraseList(pHighScores.pHSList);
    if (HSHand)
        MemHandleUnlock(HSHand);
    rec = CtlGetLabel(pHighScores.pHSGameTime)[0]-'2';
    if (CtlGetValue(pHighScores.pHSUniqueScoring)) {
	    rec = rec*8+8;
		rec += num*2;
		if (CtlGetValue(pHighScores.pHSLinearScoring))
			rec += 1;
	}
	else {
	    rec = rec*2;
	    if (CtlGetValue(pHighScores.pHSLinearScoring))
		    rec += 1;
	}
    HSHand = DmQueryRecord(highscoreref,rec);
    ErrFatalDisplayIf(!HSHand,"Error reading High Scores");
    HSPtr = MemHandleLock(HSHand);
    ErrFatalDisplayIf(!HSPtr,"Error reading High Scores");
    LstSetListChoices(pHighScores.pHSList,dummystrings,10);
    LstMakeItemVisible(pHighScores.pHSList,0);
    LstDrawList(pHighScores.pHSList);
}

void drawHighScores(UInt idx, RectanglePtr rect, CharPtr *text) {
    char scr[6];
    char *t;
    int i,j;
    UInt num;
    tHighScoreItem *item;

    num = ((tHighScoreHdr *)HSPtr)->numscores;
    item = (tHighScoreItem *)(HSPtr+sizeof(tHighScoreHdr));
    WinSetClip(rect);
    WinEraseRectangle(rect,0);
    if (idx >= num) {
        WinResetClip();
        return;
    }
    StrIToA(scr,item[idx].score);
    WinDrawChars(scr,StrLen(scr),rect->topLeft.x,rect->topLeft.y);
    WinDrawChars(item[idx].name,StrLen(item[idx].name),rect->topLeft.x+25,rect->topLeft.y);
    WinDrawChars(item[idx].date,StrLen(item[idx].date),rect->topLeft.x+95,rect->topLeft.y);
    WinResetClip();
}

int HighScoresHandleEvent(EventType *e) {
    int i;
    char *ptr;
    switch (e->eType) {
        case ctlSelectEvent:
            switch (e->data.ctlSelect.controlID) {
                case HSClear:
                    if (FrmAlert(ClearAreYouSure) == 0) {
                        UpdateHighScores(true);
                        DisplayHighScores();
                    }
                    return true;
                case HSClose:
                    if (HSHand) {
                        MemHandleUnlock(HSHand);
                        HSHand = 0;
                    }
					startup = 0;
                    FrmGotoForm(WordBox);
                    FrmUpdateForm(WordBox,frmRedrawUpdateCode);
                    return true;
                case HSLinearScoring:
                    DisplayHighScores();
                    return true;
                case HSExpScoring:
                    DisplayHighScores();
                    return true;
                case HSAllScoring:
                    DisplayHighScores();
                    return true;
                case HSUniqueScoring:
                    DisplayHighScores();
                    return true;
                case OnePlayer:
                    DisplayHighScores();
                    return true;
				case TwoPlayers:
                    DisplayHighScores();
                    return true;
				case ThreePlayers:
                    DisplayHighScores();
                    return true;
				case FourPlayers:
                    DisplayHighScores();
                    return true;
				default:
            }
            break;
        case popSelectEvent:
            FrmHandleEvent(FrmGetActiveForm(), e);
            DisplayHighScores();
            return true;
        default:
    }
    return false;
}

int WordsHandleEvent(EventType *e) {
    switch (e->eType) {
        case ctlSelectEvent:
            if (e->data.ctlSelect.controlID == Close) {
//                GameOver();
                gameinprogress = 0;
                pausemode = 0;
				startup = 0;
                FrmGotoForm(WordBox);
                FrmUpdateForm(WordBox,frmRedrawUpdateCode);
                return true;
            }
            if (e->data.ctlSelect.controlID == HSButton) {
                FrmGotoForm(HighScores);
                return true;
            }
            break;
        case popSelectEvent: {
            if (e->data.ctlSelect.controlID == PlayerScore) {
                curplayer = e->data.popSelect.selection;
                StrIToA(curscore,scores[curplayer]);
                FldSetTextPtr(pWords.pPlayerScore,curscore);
                FldEraseField(pWords.pPlayerScore);
                FldDrawField(pWords.pPlayerScore);
            }
            else {
                curplayer = e->data.popSelect.selection;
                StrIToA(curscore2,scores[curplayer]);
                FldSetTextPtr(pWords.pPlayerScore2,curscore2);
                FldEraseField(pWords.pPlayerScore2);
                FldDrawField(pWords.pPlayerScore2);
            }
			BuildWordList();
            break;
        }
        default:
    }
    return false;
}

int NameEntryHandleEvent(EventType *e) {
    int i;
    char *ptr;
    switch (e->eType) {
        case ctlSelectEvent:
            switch (e->data.ctlSelect.controlID) {
                case NewOK:
                    for (i=0;i<numplayers;i++) {
                        switch (i) {
                            case 0:
                                ptr = FldGetTextPtr(pNameEntry.pNewPlayer1);
                                if (!ptr || !*ptr) {
                                    ptr = CtlGetLabel(pNameEntry.pPlayer1);
                                }
                                break;
                            case 1:
                                ptr = FldGetTextPtr(pNameEntry.pNewPlayer2);
                                if (!ptr || !*ptr)
                                    ptr = CtlGetLabel(pNameEntry.pPlayer2);
                                break;
                            case 2:
                                ptr = FldGetTextPtr(pNameEntry.pNewPlayer3);
                                if (!ptr || !*ptr)
                                    ptr = CtlGetLabel(pNameEntry.pPlayer3);
                                break;
                            case 3:
                                ptr = FldGetTextPtr(pNameEntry.pNewPlayer4);
                                if (!ptr || !*ptr)
                                    ptr = CtlGetLabel(pNameEntry.pPlayer4);
                                break;
                        }
                        if (ptr) {
                            StrCopy(curplayername[i],ptr);
                            ArchivePlayer(curplayername[i],i);
                        }
                    }
					startup = 1;
                    gametime = (*CtlGetLabel(pNameEntry.pGameTime)-'0')*60;
                    FrmGotoForm(WordBox);
                    FrmUpdateForm(WordBox,frmRedrawUpdateCode);
                    return true;
                case NewCancel:
					startup = 0;
                    FrmGotoForm(WordBox);
                    FrmUpdateForm(WordBox,frmRedrawUpdateCode);
                    return true;
                case OnePlayer:
                    SetNumPlayers(1);
                    return true;
                case TwoPlayers:
                    SetNumPlayers(2);
                    return true;
                case ThreePlayers:
                    SetNumPlayers(3);
                    return true;
                case FourPlayers:
                    SetNumPlayers(4);
                    return true;
                case LinearScoring:
                    prefptr->linearscoring = 1;
                    return true;
                case ExpScoring:
                    prefptr->linearscoring = 0;
                    return true;
                case AllScoring:
                    prefptr->uniquescoring = 0;
                    return true;
                case UniqueScoring:
                    prefptr->uniquescoring = 1;
                    return true;
                default:
            }
            break;
        default:
    }
    return false;
}

int WordBoxHandleEvent(EventType *e) {
    int i,rc,l,j;
    long ts;

    switch (e->eType) {
        case nilEvent:
            UpdateClock();
            return true;
        case menuEvent:
	    	switch (e->data.menu.itemID) {
		    	case MenuNew:
	    			if (gameinprogress)
						if (!FrmAlert(NewGameConf)) {
					  	    starttime = TimGetTicks();
							return false;
						}
  		    	    GameOver();
                    FrmGotoForm(NameEntry);
		    	    return true;
                case MenuReplaylast:
                    replaylast = true;
                    GameOver();
					if (flags[0])
						LoadSavedGame(true);
					for (i=0;i<25;i++)
						flags[i] = 0;
                    FrmGotoForm(NameEntry);
                    return true;
		    	case MenuHighscores:
		    	    FrmGotoForm(HighScores);
		    	    return true;
		    	case MenuAbout:
		    	    FrmAlert(HelpAlert);
		    	    return true;
		    	case MenuRules:
		    	    FrmAlert(RulesAlert);
		    	    return true;
    		}
            break;
        case frmUpdateEvent:
            FrmDrawForm(FrmGetActiveForm());
            FldEraseField(pWordBox.pScore);
            FldDrawField(pWordBox.pScore);
            FldEraseField(pWordBox.pTimeLeft);
            FldDrawField(pWordBox.pTimeLeft);
            DrawBoard();
            return true;
        case penDownEvent:
            if (pausemode) {
                return false;
            }
            for (i=0;i<25;i++)
                flags[i] = 0;
            startx = -1;
            starty = -1;
            index = 0;
            processStroke(e);
            FrmHandleEvent(FrmGetActiveForm(), e);
            return true;
        case penUpEvent:
            if ((!gameinprogress) || pausemode || !index) {
                return true;
            }
            for (i=0;i<25;i++)
                flags[i] = 0;
            cliprect.extent.x = 0;
            FrmUpdateForm(WordBox,frmRedrawUpdateCode);
            lookup[index] = 0;
            l = StrLen(lookup);
            for (i=0;i<l;i++) {
                if (lookup[i] == 'Q') {
                    for (j=l;j>i;j--)
                        lookup[j+1] = lookup[j];
                    lookup[j+1] = 'U';
                    l++;
                    index++;
                }
            }
            if (index >= 4) {
                lookup[index] = 0;
                StrToLower(lookup,lookup);
                rc = VerifyWord(lookup);
                if (rc == 1) {
                    *lastword = 0;
                    if (prefptr->linearscoring)
                        ts = linearscorevalues[StrLen(lookup)-4];
                    else
                        ts = expscorevalues[StrLen(lookup)-4];
                    StrCat(lastword,lookup);
                    StrIToA(lookup,ts);
                    StrCat(lastword," ");
                    StrCat(lastword,lookup);
                    FldSetTextPtr(pWordBox.pLastWord,lastword);
                    FldEraseField(pWordBox.pLastWord);
                    FldDrawField(pWordBox.pLastWord);
                    scores[curplayer] += ts;
                    StrIToA(curscore,scores[curplayer]);
                    FldSetTextPtr(pWordBox.pScore,curscore);
                    FldEraseField(pWordBox.pScore);
                    FldDrawField(pWordBox.pScore);
                    goodwordcount ++;
                }
                else {
                    *lastword = 0;
                    StrCat(lastword,lookup);
                    if (rc == 0) {
                        badwordcount++;
                        if (badwordcount >= goodwordcount) {
                            SndDoCmd(0,&sndcmd3,false);
                            StrCat(lastword," TOT!");
                        }
                        else {
                            SndDoCmd(0,&sndcmd,false);
                            StrCat(lastword," BAD");
                        }
                    }
                    else {
                        SndDoCmd(0,&sndcmd,false);
                        StrCat(lastword," DUP");
                    }
                    FldSetTextPtr(pWordBox.pLastWord,lastword);
                    FldEraseField(pWordBox.pLastWord);
                    FldDrawField(pWordBox.pLastWord);
                }
            }
            else {
                SndDoCmd(0,&sndcmd,false);
                *lastword = 0;
                StrToLower(lookup,lookup);
                StrCat(lastword,lookup);
                StrCat(lastword," LEN");
                FldSetTextPtr(pWordBox.pLastWord,lastword);
                FldEraseField(pWordBox.pLastWord);
                FldDrawField(pWordBox.pLastWord);
            }
            FrmHandleEvent(FrmGetActiveForm(), e);
            UpdateClock();
            return true;
        case penMoveEvent:
            processStroke(e);
            FrmHandleEvent(FrmGetActiveForm(), e);
            UpdateClock();
            return true;
        case ctlSelectEvent:
            if (e->data.ctlSelect.controlID == Pause) {
                if (pausemode) {
                    starttime = TimGetTicks();
                    pausemode = 0;
                }
                else {
                    cliprect.extent.x = 0;
                    pausemode = 1;
                }
                FrmUpdateForm(WordBox,frmRedrawUpdateCode);
            }
            else if (e->data.ctlSelect.controlID == WordsButton) {
                if (gameinprogress || !wordlistptr[0])
                    SndDoCmd(0,&sndcmd,false);
                else
                    FrmGotoForm(Words);
                return true;
            }
            else if (e->data.ctlSelect.controlID == Cancel) {
                if (FrmAlert(CancelAreYouSure) == 0) {
                    pausemode = 0;
                    gametimeleft = 0;
                    UpdateClock();
                }
                return true;
            }
            else if (e->data.ctlSelect.controlID == NewButton) {
	    	    if (gameinprogress)
					if (FrmAlert(NewGameConf) != 0) {
				  	    starttime = TimGetTicks();
						return false;
					}
				GameOver();
                FrmGotoForm(NameEntry);
	    	    return true;
            }
            break;
    }
    return false;
}

int FormHandleEvent(EventType *e) {
    switch (FrmGetActiveFormID()) {
        case WordBox:
            return WordBoxHandleEvent(e);
        case Words:
            return WordsHandleEvent(e);
        case NameEntry:
            return NameEntryHandleEvent(e);
        case HighScores:
            return HighScoresHandleEvent(e);
        default:
    }
    return false;
}

void drawWord(UInt idx, RectanglePtr rect, CharPtr *text) {
    char scr[6];
    char *t;
    int i,j;

    WinSetClip(rect);
    WinEraseRectangle(rect,0);
    j=0;
    t = text[idx];
    for (i=1;i<5;i++)
        if (t[i] != ' ') {
            scr[j] = t[i];
            j++;
        }
    scr[j] = 0;
    WinDrawChars(scr,StrLen(scr),rect->topLeft.x,rect->topLeft.y);
    WinDrawChars(t+7,StrLen(t+7),rect->topLeft.x+20,rect->topLeft.y);
    WinDrawChars(t,1,rect->topLeft.x+14,rect->topLeft.y);
    WinResetClip();
}

void LoadFormVars(Word id, FormPtr form) {
    int i;

    switch (id) {
        case WordBox:
            pWordBox.pTimeLeft = FrmGetObjectPtr(form,FrmGetObjectIndex(form,TimeLeft));
            pWordBox.pPause = FrmGetObjectPtr(form,FrmGetObjectIndex(form,Pause));
            pWordBox.pLastWord = FrmGetObjectPtr(form,FrmGetObjectIndex(form,LastWord));
            pWordBox.pScore = FrmGetObjectPtr(form,FrmGetObjectIndex(form,Score));
            pWordBox.pCancel = FrmGetObjectPtr(form,FrmGetObjectIndex(form,Cancel));
            pWordBox.pAccept = FrmGetObjectPtr(form,FrmGetObjectIndex(form,Accept));
            pWordBox.pName = FrmGetObjectPtr(form,FrmGetObjectIndex(form,Name));
            pWordBox.pWordsButton = FrmGetObjectPtr(form,FrmGetObjectIndex(form,WordsButton));
            pWordBox.pNewButton = FrmGetObjectPtr(form,FrmGetObjectIndex(form,NewButton));
            break;
        case Words:
            pWords.pClose = FrmGetObjectPtr(form,FrmGetObjectIndex(form,Close));
            pWords.pPlayerNames = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PlayerNames));
            pWords.pPlayerScore = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PlayerScore));
            pWords.pPlayerWords = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PlayerWords));
            pWords.pPlayerNameList = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PilotList21));
            pWords.pPlayerNames2 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PlayerNames2));
            pWords.pPlayerScore2 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PlayerScore2));
            pWords.pPlayerWords2 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PlayerWords2));
            pWords.pPlayerName2List = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PlayerNames2List));
            pWords.pHSButton = FrmGetObjectPtr(form,FrmGetObjectIndex(form,HSButton));
            pWords.pOptions = FrmGetObjectPtr(form,FrmGetObjectIndex(form,Options));
            break;
        case NameEntry:
            pNameEntry.pOnePlayer = FrmGetObjectPtr(form,FrmGetObjectIndex(form,OnePlayer));
            pNameEntry.pTwoPlayers = FrmGetObjectPtr(form,FrmGetObjectIndex(form,TwoPlayers));
            pNameEntry.pThreePlayers = FrmGetObjectPtr(form,FrmGetObjectIndex(form,ThreePlayers));
            pNameEntry.pFourPlayers = FrmGetObjectPtr(form,FrmGetObjectIndex(form,FourPlayers));
            pNameEntry.pPlayer1 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,Player1));
            pNameEntry.pPlayer2 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,Player2));
            pNameEntry.pPlayer3 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,Player3));
            pNameEntry.pPlayer4 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,Player4));
            pNameEntry.pNewPlayer1 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,NewPlayer1));
            pNameEntry.pNewPlayer2 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,NewPlayer2));
            pNameEntry.pNewPlayer3 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,NewPlayer3));
            pNameEntry.pNewPlayer4 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,NewPlayer4));
            pNameEntry.pPilotList41 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PilotList41));
            pNameEntry.pPilotList42 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PilotList42));
            pNameEntry.pPilotList43 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PilotList43));
            pNameEntry.pPilotList44 = FrmGetObjectPtr(form,FrmGetObjectIndex(form,PilotList44));
            pNameEntry.pNewOK = FrmGetObjectPtr(form,FrmGetObjectIndex(form,NewOK));
            pNameEntry.pNewCancel = FrmGetObjectPtr(form,FrmGetObjectIndex(form,NewCancel));
            pNameEntry.pGameTime = FrmGetObjectPtr(form,FrmGetObjectIndex(form,GameTime));
            pNameEntry.pLinearScoring = FrmGetObjectPtr(form,FrmGetObjectIndex(form,LinearScoring));
            pNameEntry.pExpScoring = FrmGetObjectPtr(form,FrmGetObjectIndex(form,ExpScoring));
            pNameEntry.pAllScoring = FrmGetObjectPtr(form,FrmGetObjectIndex(form,AllScoring));
            pNameEntry.pUniqueScoring = FrmGetObjectPtr(form,FrmGetObjectIndex(form,UniqueScoring));
            break;
        case HighScores:
            pHighScores.pHSGameTime = FrmGetObjectPtr(form,FrmGetObjectIndex(form,HSGameTime));
            pHighScores.pHSGameTimeList = FrmGetObjectPtr(form,FrmGetObjectIndex(form,HSGameTimeList));
            pHighScores.pHSExpScoring = FrmGetObjectPtr(form,FrmGetObjectIndex(form,HSExpScoring));
            pHighScores.pHSLinearScoring = FrmGetObjectPtr(form,FrmGetObjectIndex(form,HSLinearScoring));
            pHighScores.pHSAllScoring = FrmGetObjectPtr(form,FrmGetObjectIndex(form,HSAllScoring));
            pHighScores.pHSUniqueScoring = FrmGetObjectPtr(form,FrmGetObjectIndex(form,HSUniqueScoring));
            pHighScores.pHSList = FrmGetObjectPtr(form,FrmGetObjectIndex(form,HSList));
            pHighScores.pHSClose = FrmGetObjectPtr(form,FrmGetObjectIndex(form,HSClose));
            pHighScores.pHSClear = FrmGetObjectPtr(form,FrmGetObjectIndex(form,HSClear));
            pHighScores.pHSGameTime = FrmGetObjectPtr(form,FrmGetObjectIndex(form,HSGameTime));
            pHighScores.pOnePlayer = FrmGetObjectPtr(form,FrmGetObjectIndex(form,OnePlayer));
            pHighScores.pTwoPlayers = FrmGetObjectPtr(form,FrmGetObjectIndex(form,TwoPlayers));
            pHighScores.pThreePlayers = FrmGetObjectPtr(form,FrmGetObjectIndex(form,ThreePlayers));
            pHighScores.pFourPlayers = FrmGetObjectPtr(form,FrmGetObjectIndex(form,FourPlayers));
            break;
        default:
    }
}

void InitForm(Word id, FormPtr form) {
    int i,j;

    switch (id) {
        case WordBox:
			StrIToA(curscore,scores[curplayer]);
			StrCopy(lastword,"");
            FldSetTextPtr(pWordBox.pName,curplayername[curplayer]);
            FldSetTextPtr(pWordBox.pTimeLeft,curtime);
            FldSetTextPtr(pWordBox.pScore,curscore);
            *lastword = 0;
            FldSetTextPtr(pWordBox.pLastWord,lastword);
//            curplayer = 0;
//            FldSetTextPtr(pWordBox.pName,curplayername[curplayer]);
    	    starttime = TimGetTicks();
			if (startup) {
				if (startup == 1) {
					GenerateGame();
	    			gametimeleft = gametime*100;
					StrCopy(curscore,"0");
		            FldSetTextPtr(pWordBox.pScore,curscore);
					curplayer = 0;
		            FldSetTextPtr(pWordBox.pName,curplayername[curplayer]);
    				for (i=0;i<numplayers;i++) {
						scores[i] = 0;
						if (wordlisthand[i] == 0) {
							wordlisthand[i] = MemHandleNew(4000);
							ErrFatalDisplayIf(!wordlisthand[i],"Could not allocate words memory");
    	    				wordlistptr[i] = MemHandleLock(wordlisthand[i]);
						}
						wordlistwalk[i] = wordlistptr[i];
    	    			*(wordlistwalk[i]) = 0;
        			}
				}
				starttime = TimGetTicks();
				pausemode = 0;
				gameinprogress = 1;
				UpdateClock();
    			pausemode = 1;
			}
			else {
				gameinprogress = 0;
				pausemode = 0;
				startup = 1;
			}
            break;
        case Words:
            if (prefptr->linearscoring)
				if (prefptr->uniquescoring)
					CtlSetLabel(pWords.pOptions,"Linear,Unique");
				else
					CtlSetLabel(pWords.pOptions,"Linear,All");
			else
				if (prefptr->uniquescoring)
					CtlSetLabel(pWords.pOptions,"Exp,Unique");
				else
					CtlSetLabel(pWords.pOptions,"Exp,All");

            LstSetDrawFunction(pWords.pPlayerWords,(ListDrawDataFuncType *)drawWord);
            LstSetDrawFunction(pWords.pPlayerWords2,(ListDrawDataFuncType *)drawWord);
            StrIToA(curscore,scores[0]);
            FldSetTextPtr(pWords.pPlayerScore,curscore);
            for (i=0;i<numplayers;i++)
                dumbptrs[i] = curplayername[i];
            CtlSetLabel(pWords.pPlayerNames,dumbptrs[0]);
            if (numplayers < 4)
				LstSetHeight(pWords.pPlayerNameList,numplayers);
			LstSetListChoices(pWords.pPlayerNameList,dumbptrs,numplayers);
            LstSetSelection(pWords.pPlayerNameList,0);
            if (numplayers < 4)
				LstSetHeight(pWords.pPlayerName2List,numplayers);
            LstSetListChoices(pWords.pPlayerName2List,dumbptrs,numplayers);
            if (numplayers > 1) {
                CtlSetLabel(pWords.pPlayerNames2,dumbptrs[1]);
                LstSetSelection(pWords.pPlayerName2List,1);
                StrIToA(curscore2,scores[1]);
                FldSetTextPtr(pWords.pPlayerScore2,curscore2);
            }
            else {
                CtlSetLabel(pWords.pPlayerNames2,dumbptrs[0]);
                LstSetSelection(pWords.pPlayerName2List,0);
            }
            BuildWordList();
            break;
        case NameEntry:
			GrfSetState(false,false,true);
            if (prefptr->linearscoring)
                CtlSetValue(pNameEntry.pLinearScoring,1);
            else
                CtlSetValue(pNameEntry.pExpScoring,1);
            if (prefptr->uniquescoring)
                CtlSetValue(pNameEntry.pUniqueScoring,1);
            else
                CtlSetValue(pNameEntry.pAllScoring,1);
            switch (gametime) {
                case 300:
                    CtlSetLabel(pNameEntry.pGameTime,"5:00");
                    break;
                case 240:
                    CtlSetLabel(pNameEntry.pGameTime,"4:00");
                    break;
                case 180:
                    CtlSetLabel(pNameEntry.pGameTime,"3:00");
                    break;
                case 120:
                    CtlSetLabel(pNameEntry.pGameTime,"2:00");
                    break;
            }
            SetNumPlayers(numplayers);
//            for (i=0;i<historyplayers;i++) {
//                history[i] = prefptr->names[i];
//            }
			if (historyplayers) {
				if (PLists)
					for (j=0;j<4;j++)
						for (i=0;i<historyplayers;i++) {
							(*PPtrs)[j][i] = prefptr->names[prefptr->archiveorder[j][i]];
						}
				if (historyplayers < 4)
					LstSetHeight(pNameEntry.pPilotList41,historyplayers);
				LstSetListChoices(pNameEntry.pPilotList41,(*PPtrs)[0],historyplayers);
//				LstSetListChoices(pNameEntry.pPilotList41,history,historyplayers);
                if (historyplayers < 4)
					LstSetHeight(pNameEntry.pPilotList42,historyplayers);
                LstSetListChoices(pNameEntry.pPilotList42,(*PPtrs)[1],historyplayers);
//                LstSetListChoices(pNameEntry.pPilotList42,history,historyplayers);
                if (historyplayers < 4)
					LstSetHeight(pNameEntry.pPilotList43,historyplayers);
                LstSetListChoices(pNameEntry.pPilotList43,(*PPtrs)[2],historyplayers);
//                LstSetListChoices(pNameEntry.pPilotList43,history,historyplayers);
                if (historyplayers < 4)
					LstSetHeight(pNameEntry.pPilotList44,historyplayers);
                LstSetListChoices(pNameEntry.pPilotList44,(*PPtrs)[3],historyplayers);
//                LstSetListChoices(pNameEntry.pPilotList44,history,historyplayers);
                CtlSetLabel(pNameEntry.pPlayer1,prefptr->names[prefptr->archiveorder[0][0]]);
                CtlSetLabel(pNameEntry.pPlayer2,prefptr->names[prefptr->archiveorder[1][0]]);
                CtlSetLabel(pNameEntry.pPlayer3,prefptr->names[prefptr->archiveorder[2][0]]);
                CtlSetLabel(pNameEntry.pPlayer4,prefptr->names[prefptr->archiveorder[3][0]]);
            }
            break;
        case HighScores:
            LstSetDrawFunction(pHighScores.pHSList,(ListDrawDataFuncType *)drawHighScores);
            if (prefptr->linearscoring)
                CtlSetValue(pHighScores.pHSLinearScoring,1);
            else
                CtlSetValue(pHighScores.pHSExpScoring,1);
            if (prefptr->uniquescoring)
                CtlSetValue(pHighScores.pHSUniqueScoring,1);
            else
                CtlSetValue(pHighScores.pHSAllScoring,1);
            switch (gametime) {
                case 300:
                    CtlSetLabel(pHighScores.pHSGameTime,"5:00");
                    break;
                case 240:
                    CtlSetLabel(pHighScores.pHSGameTime,"4:00");
                    break;
                case 180:
                    CtlSetLabel(pHighScores.pHSGameTime,"3:00");
                    break;
                case 120:
                    CtlSetLabel(pHighScores.pHSGameTime,"2:00");
                    break;
            }
			switch (numplayers) {
				case 1:
		            CtlSetValue(pHighScores.pOnePlayer,1);
					break;
				case 2:
		            CtlSetValue(pHighScores.pTwoPlayers,1);
					break;
				case 3:
		            CtlSetValue(pHighScores.pThreePlayers,1);
					break;
				default:
		            CtlSetValue(pHighScores.pFourPlayers,1);
					break;
			}
            DisplayHighScores();
            break;
        default:
    }
}

